package gaborFilter;


import icy.image.IcyBufferedImage;
import icy.main.Icy;
import icy.sequence.Sequence;
import icy.type.DataType;
import ij.*;
import ij.process.*;
import ij.plugin.filter.*;
import ij.plugin.ContrastEnhancer;
import ij.ImagePlus;


/**
* This script calculates a set of Gabor filters over the selected image.
*
* Parameters: sigma, gamma, psi, Fx, nAngles
*/

public class GaborFilter {
	
	 
	// Sigma defining the size of the Gaussian envelope
	private static double sigma = 2.0;//8.0
	// Aspect ratio of the Gaussian curves
	private static double gamma = 0.05;//0.25;
	// Phase
	private static double psi = Math.PI / 4.0 * 0;
	// Frequency of the sinusoidal component
	private static double Fx = 1.0;//3.0;
	 
	// Number of diferent orientation angles to use
	private static int nAngles = 15;//5;
	
	//image
	private static ImagePlus originalImage=null; 
	
	public GaborFilter(IcyBufferedImage image) {
		IcyBufferedImage im=image;
		originalImage=new ImagePlus("image",im);
	}

	// ----------------------------------------------------------------------------
	public GaborFilter(double[] image, int width, int height) {
		
		IcyBufferedImage im=convertDouble1DToIcyBufferedImage(image, width, height);
		originalImage=new ImagePlus("image",im);
		//originalImage.show();
	}
	
	// ----------------------------------------------------------------------------
	public int getNbAngles(){
		return nAngles;
	}
	
	// ----------------------------------------------------------------------------
	private IcyBufferedImage convertDouble1DToIcyBufferedImage(double[] image, int width, int height){	
		
		IcyBufferedImage tempImage = new IcyBufferedImage(width, height, 1, DataType.DOUBLE);
		double[] temp = tempImage.getDataXYAsDouble(0);
			
		for (int y = 0; y < height; y++) {
			for (int x = 0; x < width; x++) {

				temp[x + y * width] = image[x + y * width];

			}
		}
			
		tempImage.dataChanged();
		tempImage.updateChannelsBounds();
		
		return tempImage;
	}
	
	// ----------------------------------------------------------------------------
	public ImageStack run(){
	// copy original image and transform it to 32 bit 
	//ImagePlus originalImage = new ImagePlus("/Users/anais/Desktop/leaf-1.tif");
	originalImage = new ImagePlus(originalImage.getTitle(), originalImage.getProcessor().convertToFloat());
	int width = originalImage.getWidth();
	int height = originalImage.getHeight();
	 
	// Apply aspect ratio to the Gaussian curves
	double sigma_x = sigma;
	double sigma_y = sigma / gamma;
	 
	// Decide size of the filters based on the sigma
	double largerSigma = (sigma_x > sigma_y) ? (int) sigma_x : (int) sigma_y;
	if(largerSigma < 1) largerSigma = 1;
	     
	ImageProcessor ip = originalImage.getProcessor().duplicate();
	 
	double sigma_x2 = sigma_x * sigma_x;
	double sigma_y2 = sigma_y * sigma_y;
 
	// Create set of filters
	 
	int filterSizeX = 19; //6 * largerSigma + 1;
	int filterSizeY = 19; //6 * largerSigma + 1;

	
	int middleX = (int) Math.round(filterSizeX / 2);
	int middleY = (int) Math.round(filterSizeY / 2);
 
	ImageStack is = new ImageStack(width, height);
	ImageStack kernels = new ImageStack(filterSizeX, filterSizeY);
 
	double rotationAngle = Math.PI/(double)nAngles;
	
	// Rotate kernel from 0 to 180 degrees
	
	for (int i=0; i<nAngles; i++)
	{   
		double theta = rotationAngle * i;
	    FloatProcessor filter = new FloatProcessor(filterSizeX, filterSizeY);  
	    for (int x=-middleX; x<=middleX; x++)
	    {
	        for (int y=-middleY; y<=middleY; y++)
	        {           
	            double xPrime = (double)x * Math.cos(theta) + (double)y * Math.sin(theta);
	            double yPrime = (double)y * Math.cos(theta) - (double)x * Math.sin(theta);
	                 
	            double a = 1.0 / ( 2.0 * Math.PI * sigma_x * sigma_y ) *
	                            Math.exp(-0.5 * (xPrime*xPrime / sigma_x2 + yPrime*yPrime / sigma_y2) );
	            double c = Math.cos( 2.0 * Math.PI * (Fx * xPrime) / filterSizeX + psi); 
	             
	            filter.setf(x+middleX, y+middleY, (float)(a*c) );
	        }
	    }
	    kernels.addSlice("kernel angle = " + theta, filter);
	}
	 
	// Show kernels
//	ImagePlus ip_kernels = new ImagePlus("kernels", kernels);
//	ip_kernels.show();
	 
	// Apply kernels
	for (int i=0; i<nAngles; i++)
	{
	    double theta = rotationAngle * i;      
	    Convolver c = new Convolver();                
	     
	    float[] kernel = (float[]) kernels.getProcessor(i+1).getPixels();
	    ip = originalImage.getProcessor().duplicate();      
	    c.convolveFloat(ip, kernel, filterSizeX, filterSizeY);      
	 
	    is.addSlice("gabor angle = " + i, ip);
	}
	                             
	 
	// Normalize filtered stack
	ContrastEnhancer c = new ContrastEnhancer();
	for(int i=1 ; i <= is.getSize(); i++)
	{
	    c.stretchHistogram(is.getProcessor(i), 0.4);
	}

	ImagePlus projectStack = new ImagePlus("filtered stack",is);
	IJ.run(projectStack, "Enhance Contrast", "saturated=0.4 normalize normalize_all");
	
	//Ajout To update the calibration table
	is.update(projectStack.getProcessor());
	//fin ajout
	
	//(new ImagePlus("gabor, sigma="+sigma+" gamma="+gamma+ " psi="+psi, is)).show();
	 
	return is;
	
	}
}
