/*******************************************************************************
 * Copyright (c) 2012-2013 Biomedical Image Group (BIG), EPFL, Switzerland.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Public License v3.0
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/gpl.html
 * 
 * Contributors:
 *     Ricard Delgado-Gonzalo (ricard.delgado@gmail.com)
 *     Nicolas Chenouard (nicolas.chenouard@gmail.com)
 *     Philippe Th&#233;venaz (philippe.thevenaz@epfl.ch)
 *     Emrah Bostan (emrah.bostan@gmail.com)
 *     Ulugbek S. Kamilov (kamilov@gmail.com)
 *     Ramtin Madani (ramtin_madani@yahoo.com)
 *     Masih Nilchian (masih_n85@yahoo.com)
 *     C&#233;dric Vonesch (cedric.vonesch@epfl.ch)
 *     Virginie Uhlmann (virginie.uhlmann@epfl.ch)
 *     Cl&#233;ment Marti (clement.marti@epfl.ch)
 *     Julien Jacquemot (julien.jacquemot@epfl.ch)
 ******************************************************************************/
package plugins.big.texturesnake.core;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.Arrays;
import java.util.Scanner;
import java.util.Vector;

import plugins.kernel.roi.roi2d.ROI2DRectangle;
import icy.common.exception.UnsupportedFormatException;
import icy.file.Loader;
import icy.file.SequenceFileImporter;
import icy.gui.frame.progress.ProgressFrame;
import icy.image.IcyBufferedImage;
import icy.image.IcyBufferedImageUtil;
import icy.image.ImageUtil;
import icy.main.Icy;
import icy.sequence.Sequence;
import icy.sequence.SequenceUtil;
import icy.type.DataType;
import ij.ImagePlus;
import ij.ImageStack;
import ij.io.FileSaver;
import ij.io.Opener;
import ij.measure.Calibration;
import ij.plugin.RGBStackMerge;
import ij.process.ByteProcessor;
import ij.process.ColorProcessor;
import ij.process.FloatProcessor;
import ij.process.ImageProcessor;
import ij.process.ShortProcessor;

import Jama.EigenvalueDecomposition;
import Jama.Matrix;
import fftacademic.AcademicFFT;
import gaborFilter.GaborFilter;
import waveletpackage.AnalysisR;

/**
 * Container for the LUTs.
 * 
 * @version January 22, 2017
 * 
 * @author Anaïs Badoual (anais.badoual@epfl.ch)
 * @author Yann Perret (yann.perret@epfl.ch)
 */
public class ImageLUTContainer {

	// ----------------------------------------------------------------------------
	// IMAGE

	/** <code>Sequence</code> of interest. */
	private static Sequence sequence_ = null;
	/** Array containing the image data of the active channel. */
	private double[][] imageDataArray_ = null;
	/**
	 * Array containing the preintegrated image data along the vertical
	 * direction of the active channel.
	 */
	private double[][] preintegratedImageDataArray_ = null;
	private double[][] extendedPreintegratedImageDataArray_=null;


	// ----------------------------------------------------------------------------
	// IMAGE PROPERTIES

	/** Width of the original image. */
	private static int width_ = 0;
	/** Height of the original image. */
	private static int height_ = 0;

	// ----------------------------------------------------------------------------
	// PARAMETERS

	/** Active channel of the image. */
	private int channelNumber_ = 0;


	// ----------------------------------------------------------------------------
	// Circularly harmonic parameters

	//Set of harmonic numbers
	private static int[] harmonics_ ={ 0, 1, 2, 3, 4};//{  0};//
	//Number of scales
	private static int nscales_ =3;//64;//3;

	//Array containing the CHW decomposition of the image data for each channel
	double[][][][] chwDecomposition_ = null;

	/**
	 * Array containing the preintegrated CHW decomposition of the image data along the vertical
	 * direction of the active channel.
	 */
	double[][][][] preintegratedChwDecomposition_ = null;
	double[][][][] extendedPreintegratedChwDecomposition_ = null;

	//Correlation matrix between the response maps
	double[][] correlationMatrix_ = null;

	//Textures used for Fisher
	double[][] texture1_ = null;
	double[][] texture2_ = null;
	int textureWidth_=0;
	int textureHeight_=0;
	
	ROI2DRectangle ROITexture1_=null;
	ROI2DRectangle ROITexture2_=null;

	//Weights for each response computed with fisher discriminant method
	double[] fisherWeights_ = null;

	// ----------------------------------------------------------------------------
	// OTHER

	/** If <code>true</code>, the LUTs are consistent with the input parameters. */
	private boolean isLUTUpToDate_ = false;

	// ----------------------------------------------------------------------------
	// XML LABELS

	/**
	 * Label of the XML tag containing the information of the channel number
	 * used in the snake.
	 */
	public static final String ID_ACTIVECHANNEL = "activeChannel";

	// ============================================================================
	// PUBLIC METHODS

	/** Default constructor. */
	public ImageLUTContainer() {
		
		//Example how to open the two target texture
//		Opener opener = new Opener();
//
//		ImagePlus texture1IP = opener.openImage("/Users/Anais/switchdrive/Projets/Texture Snakes/Yann_Perret/data/Images/Texture_set1/bark1_t0.png");
//		texture1_ = ImagePlus2Array1D(texture1IP);
//
//		ImagePlus texture2IP = opener.openImage("/Users/Anais/switchdrive/Projets/Texture Snakes/Yann_Perret/data/Images/Texture_set1/carpet14_t0.png");
//		texture2_ = ImagePlus2Array1D(texture2IP);

	}

	// ----------------------------------------------------------------------------

	/** Constructs the look-up-tables that the snake needs to function. */
	/**
	 * @throws Exception
	 */
	public void buildLUTs() throws Exception {

		//ProgressFrame pFrame = new ProgressFrame("Precomputing LUTs");
		//pFrame.setLength(4 * width_);
		
		//pFrame.setPosition(0);

		//Load the R, G and B channels of the original image
		
		int nbChannel=sequence_.getSizeC();
		
		IcyBufferedImage[] inputRGBImage =new IcyBufferedImage[nbChannel];
		IcyBufferedImage[] inputCopy =new IcyBufferedImage[nbChannel];
		imageDataArray_=new double[nbChannel][];
		
		for(int i=0; i<nbChannel; i++){
			Sequence singleChannelSequence = SequenceUtil.extractChannel(sequence_,
					i);

			inputRGBImage[i]=singleChannelSequence.getFirstImage();
			
			//To display each channel
//			Sequence sequence = new Sequence("Channel " + i);
//			sequence.addImage(inputRGBImage[i]);
//			Icy.getMainInterface().addSequence(sequence);
					
			inputCopy[i] = IcyBufferedImageUtil.convertToType(
					inputRGBImage[i], DataType.DOUBLE, false);
			imageDataArray_[i] = inputCopy[i].getDataXYAsDouble(0);
			
			singleChannelSequence.close();
		}
		
		inputRGBImage=null;
		
		width_ = SequenceUtil.extractChannel(sequence_,0).getWidth();
		height_ =SequenceUtil.extractChannel(sequence_,0).getHeight();
		

		//pFrame.setPosition(width_);
		
		if(texture1_==null){
			textureWidth_=(int) (width_*0.5);
			textureHeight_=(int) (height_*0.5);
			
			int X1=0;
			int Y1=0;
			int X2=textureWidth_-1;
			int Y2=textureHeight_-1;
					
			ROITexture1_=new ROI2DRectangle(X1,Y1,textureWidth_+X1,textureHeight_+Y1);
			ROITexture2_=new ROI2DRectangle(X2,Y2,textureWidth_+X2,textureHeight_+Y2);
			
			IcyBufferedImage image=Icy.getMainInterface().getActiveImage();
			//IcyBufferedImage cropImage1 = image.getSubImageCopy(X1, Y1, textureWidth_, textureHeight_);
			IcyBufferedImage cropImage1 =IcyBufferedImageUtil.getSubImage(image, X1, Y1, textureWidth_, textureHeight_);

			//IcyBufferedImage cropImage2 = image.getSubImageCopy(X2, Y2, textureWidth_,textureHeight_);
			IcyBufferedImage cropImage2 =IcyBufferedImageUtil.getSubImage(image, X2, Y2, textureWidth_, textureHeight_);

			setTexture1(cropImage1,ROITexture1_);
			setTexture2(cropImage2,ROITexture2_);
		}
		
		double[][][][] texture1ChwDecomposition=new double[nbChannel][nscales_][harmonics_.length][textureWidth_ * textureHeight_];
		double[][][][] texture2ChwDecomposition=new double[nbChannel][nscales_][harmonics_.length][textureWidth_ * textureHeight_];
		
		if(nscales_!=0){
			//Texture analysis of the image with the CHW and Creation of the response map for every scale and harmonics
			imageCHWDecomposition();//A DECOMENTER
			
			//Ajout pour les expériences
			//imageGaborDecomposition();
			
			//imageDeapLearning();
			//
			
			//Analysis of the targeted textures for Fisher
			texture1ChwDecomposition=extractTexture1Maps();
			texture2ChwDecomposition=extractTexture2Maps();
		}
		
		//pFrame.setPosition(2*width_);
		
		
		//Compute the inverse of the correlation between the chw response maps of the image
		//correlationMatrix_ = computeInverseCorrelationMatrix(chwDecomposition_);

		//Compute the weights obtained from Fisher method
		fisherWeights_ = fisherWeights(texture1ChwDecomposition,texture2ChwDecomposition, texture1_, texture2_);

//		double sumWeights=0;
//		for (int i=0;i<fisherWeights_.length; i++){
//			sumWeights+=fisherWeights_[i];
//	}
//		
//		System.out.println("sumWeights="+sumWeights);
		
		// Region energy
//		fisherWeights_=new double[nbChannel*nscales_*harmonics_.length+nbChannel];
//		fisherWeights_[0]=1;
//		for (int i=1;i<fisherWeights_.length; i++){
//			fisherWeights_[i]=0;
//		}
		//
		
		// All weigts=1
//		fisherWeights_=new double[nbChannel*nscales_*harmonics_.length+nbChannel];
//		for (int i=0;i<fisherWeights_.length; i++){
//			fisherWeights_[i]=1;
//		}
		//
		
		//String filename="/Users/anais/switchdrive/Projets/Texture_Snakes/My_project/Experiments/fisherWeights.txt";
		//saveFisherWeights(filename);
		//readFisherWeights(filename);
		
		
		//ajout: compute the ratio of the weights for color and texture
//		double weightsColor=0;
//		double weightsTexture=0;
//		double weightsTotal=0;
//				
//		for(int i=0; i<fisherWeights_.length; i++){
//			
//			weightsTotal+=Math.abs(fisherWeights_[i]);
//			
//			boolean color=false;
//			
//			for(int j=0; j<nbChannel; j++){
//				if(i==j*(nscales_*harmonics_.length + 1)){
//					weightsColor+=Math.abs(fisherWeights_[i]);
//					color=true;
//				}
//			}
//		
//		if(color==false) weightsTexture+=Math.abs(fisherWeights_[i]);
//	}
//		
//		weightsTexture=weightsTexture/weightsTotal*100;
//
//		weightsColor=weightsColor/weightsTotal*100;
//
//		
//		
//		String filenameWeightsColor="/Users/anais/switchdrive/Projets/Texture_Snakes/My_project/Experiments/weightsColor.txt";
//		saveEstimator(filenameWeightsColor,weightsColor);
//		
//		String filenameWeightsTexture="/Users/anais/switchdrive/Projets/Texture_Snakes/My_project/Experiments/weightsTexture.txt";
//		saveEstimator(filenameWeightsTexture,weightsTexture);
		//fin ajout
		
		//pFrame.setPosition(3*width_);
		
		//Ajout: display and save somme des images pondérés
		
//		double[] im=new double[width_ * height_];
//
//		for(int i=0; i<(width_ * height_); i++){
//			for(int j=0; j<nbChannel; j++){
//				im[i] +=  fisherWeights_[j*(nscales_*harmonics_.length + 1)] * imageDataArray_[j][i];
//			}
//			
//			for(int j=0; j<nbChannel; j++){
//				for (int s = 0; s < nscales_; s++) {
//					for (int h = 0; h < harmonics_.length; h++) {
//						im[i] += fisherWeights_[j*nscales_*harmonics_.length+s * harmonics_.length + h + 1+j] *chwDecomposition_[j][s][h][i];
//					}
//				}
//			}
//		}
//		
//	//	displayImage(im,"im_nbScale"+nscales_);
//			
//	
//		String filename=sequence_.getName();
//		int indexEnd = filename.indexOf(".");
//		if(indexEnd!=-1)filename=filename.substring(0, indexEnd);
//		
//		
//		FloatProcessor fp=new FloatProcessor(width_, height_);
//
//		
//		for (int y = 0; y < height_; y++) {
//			for (int x = 0; x < width_; x++) {
//
//				fp.putPixelValue(x, y, (float) im[x+y*width_]);
//
//			}
//		}
//		
//		ImagePlus ImgNew = new ImagePlus("img",fp);
//		FileSaver saver = new FileSaver(ImgNew);
//		saver.saveAsTiff("/Users/anais/Desktop/Aggregation/Database1/"+filename+".tif");
//		ImgNew.close();
		//
		
		double fuy_val;
		
		preintegratedImageDataArray_ = new double[nbChannel][width_ * height_];
		extendedPreintegratedImageDataArray_ = new double[nbChannel][2*width_ * 2*height_];
		extendedPreintegratedChwDecomposition_ = new double[nbChannel][nscales_][harmonics_.length][2*width_ * 2*height_];
		
		for(int i=0; i<nbChannel; i++){
			for (int x = 0; x < width_; x++) {
				fuy_val = 0.0;

				for (int y = 0; y < height_; y++) {
					int index = x + width_ * y;
					fuy_val += imageDataArray_[i][index];
					preintegratedImageDataArray_[i][index] = fuy_val;
				}
			}
			
			//displayImage(preintegratedImageDataArray_[i], "preintegratedImageDataArray");
			
			//Extention of the preintegrated images
			extendedPreintegratedImageDataArray_[i]=extendPreintegratedImageAlongY(preintegratedImageDataArray_[i], imageDataArray_[i]);
			//displayExtendedImage(extendedPreintegratedImageDataArray_[i], "extendedPreintegratedImageDataArray");
			
			if(nscales_!=0){
				for(int sc=0; sc<nscales_; ++sc){
					for(int ii=0; ii<harmonics_.length; ++ii){
						extendedPreintegratedChwDecomposition_[i][sc][ii]=extendPreintegratedImageAlongY(preintegratedChwDecomposition_[i][sc][ii], chwDecomposition_[i][sc][ii]);
					}
				}
				
				//displayImage(preintegratedChwDecomposition_[i][0][0], "preintegratedChwDecomposition_["+i+"][0][0]");
			//displayExtendedImage(extendedPreintegratedChwDecomposition_[i][0][0], "extendedPreintegratedChwDecomposition_["+i+"][0][0]");
			}
		}
		
		//pFrame.close();
		isLUTUpToDate_ = true;
	}
	
	// ----------------------------------------------------------------------------
	private void saveFisherWeights(String filename) throws FileNotFoundException{
		PrintWriter writer= new PrintWriter(filename);
		
		for(int i=0; i<fisherWeights_.length; i++){
			writer.print(fisherWeights_[i]);
			writer.println();
		}
		writer.close();

	}
	
	// ----------------------------------------------------------------------------
	private void saveEstimator(String filename, double estimator) throws IOException{
		FileWriter fw = new FileWriter(filename, true);

		BufferedWriter out = new BufferedWriter(fw);

		out.write(Double.toString(estimator));
		out.newLine();
		out.close();
		
		fw.close();

	}
	
	// ----------------------------------------------------------------------------
	private void readFisherWeights(String filename){
		 Scanner scan;
		    File file = new File(filename);
		    try {
		        scan = new Scanner(file);
		        
		        int i=0;

		        while(scan.hasNextDouble())
		        {
		            fisherWeights_[i]=scan.nextDouble();
		            i++;
		        }

		    } catch (FileNotFoundException e1) {
		            e1.printStackTrace();
		    }
	}
	
	
	// ----------------------------------------------------------------------------
	private double[] extendPreintegratedImageAlongY(double[] preintegratedImage, double[] originalImage){
		
		double[] extendedImage=new double[2*width_ * 2*height_];
		
		double meanValueLeft=0;
		double meanValueRight=0;
		double meanValueBottom=0;
		double meanValueUp=0;
		
		int height2=(int) height_/2;
		int width2=(int) width_/2;

		//compute the mean value of the left and right bounds of the image
		for (int y = 0; y < height_; y++) {
			int indexLeft =  width_ * y;
			int indexRight = width_-1+ width_ * y;
				meanValueLeft+= originalImage[indexLeft];
				meanValueRight+= originalImage[indexRight];
			}
		
		meanValueLeft/=height_;
		meanValueRight/=height_;

		//compute the mean value of the upper and bottom bounds of the image	
		for (int x = 0; x < width_; x++) {
			int indexUp =  x;
			int indexBottom =  x+width_*(height_-1);
			
			meanValueUp+= originalImage[indexUp];
			meanValueBottom+= originalImage[indexBottom];

			}
		
		meanValueUp/=width_;
		meanValueBottom/=width_;
		
		//load the original image in the middle of the extended image
		for (int y = height2; y < height_+height2; y++) {
			for (int x = width2; x < width_+width2; x++) {
				int index = x + 2*width_ * y;
				int index2 = (x-width2) + width_ * (y-height2);
				extendedImage[index] = preintegratedImage[index2];
			}
		}
		
		//Fill the extended image 
		//1) Fill the columns on the left and right of the original image  (not for the entire width. Just for (width/2)-1<=x<width+(width/2)-1.)
		for (int x =0; x <width2; x++) {
			for(int y=height2; y<height_+height2; y++){
				int index = x + 2*width_ * y;
				
				if(y==height2){
					extendedImage[index]=meanValueLeft;
				}else{
					int index2 = x + 2*width_ * (y-1);
					extendedImage[index]=extendedImage[index2]+meanValueLeft;
				}
			}
		}
			
		for (int x = width_+width2; x <2*width_; x++) {
			for(int y=height2; y<height_+height2; y++){
				int index = x + 2*width_ * y;
				
				if(y==height2){
					extendedImage[index]=meanValueRight;
				}else{
					int index2 = x + 2*width_ * (y-1);
					extendedImage[index]=extendedImage[index2]+meanValueRight;
				}
			}
		}

		//2)Fill the lines below and on the top of the original image
		
		for (int x =0; x < 2*width_; x++) {
			for (int y=height2-1; y>=0; y--) {
				int index = x + 2*width_ * y;
				int index2 = x + 2*width_ * (y+1);
				extendedImage[index] = extendedImage[index2]-meanValueUp;
			}				

			for (int y =height_+height2; y < 2*height_; y++) {
				int index = x + 2*width_ * y;
				int index2 = x + 2*width_ * (y-1);
				extendedImage[index] = extendedImage[index2]+meanValueBottom;
			}
		}
		
		return extendedImage;
	}
	
	
	// ----------------------------------------------------------------------------

	private double[][][][] extractTexture1Maps(){
		int nbChannel=sequence_.getSizeC();
		double[][][][] textureChwDecomposition = new double[nbChannel][nscales_][harmonics_.length][textureWidth_ * textureHeight_];	
		
		for(int i=0; i< nbChannel; i++ ){
			for(int sc=0; sc<nscales_; ++sc){
				for(int ii=0; ii<harmonics_.length; ++ii){
					
					IcyBufferedImage im=new IcyBufferedImage(width_, height_,chwDecomposition_[i][sc][ii]);

					int width1=(int)ROITexture1_.getBounds().getWidth();
					int height1=(int)ROITexture1_.getBounds().getHeight();
					
					int X1=(int)(ROITexture1_.getPosition().getX());
					int Y1=(int)(ROITexture1_.getPosition().getY());
					
					//IcyBufferedImage cropImage1 = im.getSubImageCopy(X1, Y1, width1,height1);
					IcyBufferedImage cropImage1 =IcyBufferedImageUtil.getSubImage(im, X1, Y1, width1, height1);

					
					IcyBufferedImage cropImageCopy1 = IcyBufferedImageUtil.convertToType(
							cropImage1, DataType.DOUBLE, false);
					
					textureChwDecomposition[i][sc][ii]=cropImageCopy1.getDataXYAsDouble(0);

				}
			}
		}

		
		return textureChwDecomposition;
	}
	
	// ----------------------------------------------------------------------------

	private double[][][][] extractTexture2Maps(){
		int nbChannel=sequence_.getSizeC();
		double[][][][] textureChwDecomposition = new double[nbChannel][nscales_][harmonics_.length][textureWidth_ * textureHeight_];	
		
		for(int i=0; i<nbChannel; i++){
			for(int sc=0; sc<nscales_; ++sc){
				for(int ii=0; ii<harmonics_.length; ++ii){
					
					IcyBufferedImage im=new IcyBufferedImage(width_, height_,chwDecomposition_[i][sc][ii]);

					int width2=(int)ROITexture2_.getBounds().getWidth();
					int height2=(int)ROITexture2_.getBounds().getHeight();
					
					int X2=(int)(ROITexture2_.getPosition().getX());
					int Y2=(int)(ROITexture2_.getPosition().getY());
					
					//IcyBufferedImage cropImage2 = im.getSubImageCopy(X2, Y2, width2,height2);
					IcyBufferedImage cropImage2 =IcyBufferedImageUtil.getSubImage(im, X2, Y2, width2, height2);

					
					IcyBufferedImage cropImageCopy2 = IcyBufferedImageUtil.convertToType(
							cropImage2, DataType.DOUBLE, false);
					
					textureChwDecomposition[i][sc][ii]=cropImageCopy2.getDataXYAsDouble(0);

				}
			}
		}

		return textureChwDecomposition;
	}
	
	// ----------------------------------------------------------------------------
	/** Display the extended image in ICY */	
	private void displayExtendedImage(double[] image, String name){	
		Sequence sequence = new Sequence(name);
					
		IcyBufferedImage tempImage = new IcyBufferedImage(2*width_, 2*height_, 1, DataType.DOUBLE);
		double[] temp = tempImage.getDataXYAsDouble(0);
			
		for (int y = 0; y < 2*height_; y++) {
			for (int x = 0; x < 2*width_; x++) {

				temp[x + y * 2*width_] = image[x + y * 2*width_];

			}
		}
			
		tempImage.dataChanged();
		sequence.addImage(tempImage);
		
		Icy.getMainInterface().addSequence(sequence);
	}
	
	// ----------------------------------------------------------------------------
	/** Display the extended image in ICY */	
	private void displayImage(double[] image, String name){	
		Sequence sequence = new Sequence(name);
					
		IcyBufferedImage tempImage = new IcyBufferedImage(width_, height_, 1, DataType.DOUBLE);
		double[] temp = tempImage.getDataXYAsDouble(0);
			
		for (int y = 0; y < height_; y++) {
			for (int x = 0; x < width_; x++) {

				temp[x + y * width_] = image[x + y * width_];

			}
		}
			
		tempImage.dataChanged();
		sequence.addImage(tempImage);
		
		Icy.getMainInterface().addSequence(sequence);
	}
	
	// ----------------------------------------------------------------------------
	/** Display the texture image in ICY */	
	private void displayTextureImage(double[] image, String name){	
		Sequence sequence = new Sequence(name);
					
		IcyBufferedImage tempImage = new IcyBufferedImage(textureWidth_, textureHeight_, 1, DataType.DOUBLE);
		double[] temp = tempImage.getDataXYAsDouble(0);
			
		for (int y = 0; y < textureHeight_; y++) {
			for (int x = 0; x < textureWidth_; x++) {

				temp[x + y * textureWidth_] = image[x + y * textureWidth_];

			}
		}
			
		tempImage.dataChanged();
		sequence.addImage(tempImage);
		
		Icy.getMainInterface().addSequence(sequence);
	}
	
	// ----------------------------------------------------------------------------

	private void imageCHWDecomposition(){
		int nbChannel=sequence_.getSizeC();
		chwDecomposition_ = new double[nbChannel][nscales_][harmonics_.length][width_ * height_];
		preintegratedChwDecomposition_ = new double[nbChannel][nscales_][harmonics_.length][width_ * height_];
		
		for(int i=0; i<nbChannel; i++){
			AnalysisR analysis = new AnalysisR(imageDataArray_[i].clone(), width_, height_, harmonics_, nscales_);
			analysis.analyze();
			
			//real and imaginary part of the decomposition
			double[][][] analyzedRealImages = analysis.getWdChan_re();
			double[][][] analyzedImagImages = analysis.getWdChan_im();
			
			AcademicFFT fft = new AcademicFFT(width_, height_, 0, 0);
			
			double fuy_val;
			
			//Creation of the response map for every scale and harmonics
			for(int sc=0; sc<nscales_; ++sc){
				for(int ii=0; ii<harmonics_.length; ++ii){

					fft.inverseTransform(analyzedRealImages[sc][ii], analyzedImagImages[sc][ii], null, null);

					//Decomposition of the image
					chwDecomposition_[i][sc][ii] = absolute(analyzedRealImages[sc][ii]);
				//	displayImage(chwDecomposition_[i][sc][ii], "chwDecomposition_["+i+"]["+sc+"]["+ii+"]");
	

					for (int x = 0; x < width_; x++) {
						fuy_val = 0.0;

						for (int y = 0; y < height_; y++) {
							int index = x + width_ * y;
							fuy_val += chwDecomposition_[i][sc][ii][index];
							preintegratedChwDecomposition_[i][sc][ii][index] = fuy_val;
						}
					}
				}
			}
		}
	}
	
	// ----------------------------------------------------------------------------

	private void imageGaborDecomposition(){
		int nbChannel=sequence_.getSizeC();
		String sequenceName=sequence_.getName();
		int indexEnd = sequenceName.indexOf(".");
		
		if(indexEnd!=-1) sequenceName=sequenceName.substring(0, indexEnd);
		
		chwDecomposition_ = new double[nbChannel][nscales_][harmonics_.length][width_ * height_];
		preintegratedChwDecomposition_ = new double[nbChannel][nscales_][harmonics_.length][width_ * height_];

		String folderRooth="/Users/Anais/Desktop/GaborResponseMaps/Database1_EqualMean";

		IcyBufferedImage[][][] responsesMaps = loadImages(folderRooth + "/"+sequenceName);
		
		for(int i=0; i<nbChannel; i++){
			
			double fuy_val;
			
			//Creation of the response map for every angle
			for(int sc=0; sc<nscales_; ++sc){
				for(int ii=0; ii<harmonics_.length; ++ii){

					//Decomposition of the image
					chwDecomposition_[i][sc][ii] = ImagePlus2Array1D(new ImagePlus("image",responsesMaps[i][sc][ii]));
					//displayImage(chwDecomposition_[i][sc][ii], "chwDecomposition_["+i+"]["+sc+"]["+ii+"]");

					for (int x = 0; x < width_; x++) {
						fuy_val = 0.0;

						for (int y = 0; y < height_; y++) {
							int index = x + width_ * y;
							fuy_val += chwDecomposition_[i][sc][ii][index];
							preintegratedChwDecomposition_[i][sc][ii][index] = fuy_val;
						}
					}
				}
			}
		}
	}
	
	// ----------------------------------------------------------------------------

	private void imageDeapLearning(){
		int nbChannel=sequence_.getSizeC();
		String sequenceName=sequence_.getName();
		int indexEnd = sequenceName.indexOf(".");
		
		if(indexEnd!=-1) sequenceName=sequenceName.substring(0, indexEnd);
		
		chwDecomposition_ = new double[nbChannel][nscales_][harmonics_.length][width_ * height_];
		preintegratedChwDecomposition_ = new double[nbChannel][nscales_][harmonics_.length][width_ * height_];
		
		String folderRooth="/Users/Anais/Desktop/Result_Database2_EqualMean/ConvUp1-2_+_ReLU";

		IcyBufferedImage[][][] responsesMaps = loadImages(folderRooth + "/"+sequenceName);
		
		for(int i=0; i<nbChannel; i++){
			
			double fuy_val;
			
			//Creation of the response map for every angle
			for(int sc=0; sc<nscales_; ++sc){
				for(int ii=0; ii<harmonics_.length; ++ii){

					//Decomposition of the image
					chwDecomposition_[i][sc][ii] = ImagePlus2Array1D(new ImagePlus("image",responsesMaps[i][sc][ii]));
//					displayImage(chwDecomposition_[i][sc][ii], "chwDecomposition_["+i+"]["+sc+"]["+ii+"]");
					
					for (int x = 0; x < width_; x++) {
						fuy_val = 0.0;

						for (int y = 0; y < height_; y++) {
							int index = x + width_ * y;
							fuy_val += chwDecomposition_[i][sc][ii][index];
							preintegratedChwDecomposition_[i][sc][ii][index] = fuy_val;
						}
					}
				}
			}
		}
	}
	
	// ----------------------------------------------------------------------------

	/**
	 * Load images contained in a specific folder
	 * @param folder relative path to the folder
	 * @return images
	 */
	public static IcyBufferedImage[][][] loadImages(String folder){

		File folder_ = new File(folder);
		File[] listOfFiles = folder_.listFiles();
		Arrays.sort(listOfFiles);
		
		int nbChannel=sequence_.getSizeC();
		IcyBufferedImage[][][] imagesFinal =new IcyBufferedImage[nbChannel][nscales_][harmonics_.length];
		IcyBufferedImage[] images = new IcyBufferedImage[listOfFiles.length];

		int i =0;
		Vector<String>filenames=new Vector<String>();

		
		for (File file : listOfFiles) {
			if (file.isFile()) {
				
				String filename=folder +"/"+ file.getName();

				if (filename.endsWith("Store")==false){
					filenames.add(filename.substring(0, filename.indexOf(".")));
					
				    SequenceFileImporter importer = Loader.getSequenceFileImporter(filename, false);   
				    
				    try {
						if (importer.open(filename,0)) {
						    IcyBufferedImage img = IcyBufferedImageUtil.convertToType(importer.getImage(0, 0),DataType.DOUBLE, false);
						    images[i]=IcyBufferedImageUtil.convertToType(img, DataType.DOUBLE, false);
				   // importer.close();

						}
					} catch (UnsupportedFormatException e) {
						// TODO Auto-generated catch block
						e.printStackTrace();
					} catch (IOException e) {
						// TODO Auto-generated catch block
						e.printStackTrace();
					}
				    
					i++;
				}
				else{
					images = new IcyBufferedImage[listOfFiles.length-1];
				}
			}
		}
		
		for(int m=0; m<images.length; m++){
			String name=filenames.get(m);
			int indeStringChannel=name.lastIndexOf("channel_");;
			int indeStringScale=name.lastIndexOf("scale_");
			int indeStringOrientation=name.lastIndexOf("orientation_");
			int indexVectorChannel=Integer.parseInt(name.substring(indeStringChannel+8, indeStringChannel+9));;
//			int indexVectorScale=Integer.parseInt(name.substring(indeStringScale+6, indeStringScale+7));
			int indexVectorScale=Integer.parseInt(name.substring(indeStringScale+6, indeStringScale+10)); //FOR DEEP LEARNING ONLY
			int indexVectorOrientation=Integer.parseInt(name.substring(indeStringOrientation+12, indeStringOrientation+13));
						
			imagesFinal[indexVectorChannel-1][indexVectorScale-1][indexVectorOrientation-1]=images[m];
		}
		
		images=null;
		
		return imagesFinal;
	}
	
	// ----------------------------------------------------------------------------
	private void imageGaborDecomposition2(){
		
		//To get the number of angle
		GaborFilter filter=new GaborFilter(imageDataArray_[0].clone(), width_, height_);
		nscales_=filter.getNbAngles();
		
		//To fit the parameters of the code
		harmonics_=new int[1];
		harmonics_[0]=0;
		
		
		int nbChannel=sequence_.getSizeC();
		chwDecomposition_ = new double[nbChannel][nscales_][harmonics_.length][width_ * height_];
		preintegratedChwDecomposition_ = new double[nbChannel][nscales_][harmonics_.length][width_ * height_];
		
		for(int i=0; i<nbChannel; i++){

			filter=new GaborFilter(imageDataArray_[i].clone(), width_, height_);
			ImageStack filteredMap=filter.run();	
			
			double fuy_val;
			
			//Creation of the response map for every angle
			for(int sc=0; sc<nscales_; ++sc){
				for(int ii=0; ii<harmonics_.length; ++ii){

					//Decomposition of the image
					chwDecomposition_[i][sc][ii] = ImagePlus2Array1D(new ImagePlus(filteredMap.getSliceLabel( sc+1 ), filteredMap.getProcessor( sc+1 ) ));
					displayImage(chwDecomposition_[i][sc][ii], "chwDecomposition_["+i+"]["+sc+"]["+ii+"]");

					for (int x = 0; x < width_; x++) {
						fuy_val = 0.0;

						for (int y = 0; y < height_; y++) {
							int index = x + width_ * y;
							fuy_val += chwDecomposition_[i][sc][ii][index];
							preintegratedChwDecomposition_[i][sc][ii][index] = fuy_val;
						}
					}
				}
			}
		}
	}
	
	// ----------------------------------------------------------------------------

		/**
		 * Compute the weights with fisher discrimant method to classify two classes
		 * @param textureChwDecomposition chw decomposition of the two targeted textures
		 * @param texture1 texture 1 as 1D array
		 * @param texture2 texture 2 as 1d array
		 * @return array of weights
		 * @throws IOException 
		 */
		private double[] fisherWeights(double[][][][] texture1ChwDecomposition, double[][][][] texture2ChwDecomposition, double[][] texture1, double[][] texture2) throws IOException{
			int nbChannel=sequence_.getSizeC();
			int features = (nscales_ * harmonics_.length + 1)*nbChannel;
			//int features = (nscales_ * harmonics_.length)*nbChannel;// Pour exp: que texture
			int lenghtImageTexture1=texture1[0].length;
			int lenghtImageTexture2=texture2[0].length;

			double[][] concatTexture1ChwDecomposition = new double[features][lenghtImageTexture1];
			double[][] concatTexture2ChwDecomposition = new double[features][lenghtImageTexture2];
			
			// A commenter Pour exp: que texture
			for(int i=0; i<nbChannel; i++){
				int index=i*(nscales_* harmonics_.length+1);
				concatTexture1ChwDecomposition[index] = texture1[i];
				concatTexture2ChwDecomposition[index] = texture2[i];
				
//				displayTextureImage(concatTexture1ChwDecomposition[index], "concatTexture1ChwDecomposition["+index+"]");
//				displayTextureImage(concatTexture2ChwDecomposition[index], "concatTexture2ChwDecomposition["+index+"]");
			}

			if(nscales_!=0){
				for(int i=0; i<nbChannel; i++){
					for (int sc = 0; sc < nscales_; sc++) {
						for (int h = 0; h < harmonics_.length; h++) {
							int index=i*nscales_* harmonics_.length+sc * harmonics_.length + h + 1+i;
							//int index=i*nscales_* harmonics_.length+sc * harmonics_.length + h;// Pour exp: que texture
							concatTexture1ChwDecomposition[index] = texture1ChwDecomposition[i][sc][h];
							concatTexture2ChwDecomposition[index] = texture2ChwDecomposition[i][sc][h];
							
							//displayTextureImage(concatTexture1ChwDecomposition[index], "concatTexture1ChwDecomposition["+index+"]");
							//displayTextureImage(concatTexture2ChwDecomposition[index], "concatTexture2ChwDecomposition["+index+"]");
						}
					}
				}
			}
			
			//ajout: normalization des Tetxtures pour chaque response maps : (pixel-mean)/ecartType où mean et ecartType sont la moyenne et ecart type sur les deux textures
			
//			int n=concatTexture1ChwDecomposition[0].length;
//			double mean=0;
//			double std=0;
//			for(int f=0; f<features; f++){
//				
//				mean=0;
//				std=0;
//				
//				for(int i=0; i<n; i++){
//					mean+=concatTexture1ChwDecomposition[f][i]+concatTexture2ChwDecomposition[f][i];
//				}
//				mean=mean/(2*n);
//				
//				for(int i=0; i<n; i++){
//					std+=(concatTexture1ChwDecomposition[f][i]-mean)*(concatTexture1ChwDecomposition[f][i]-mean)+(concatTexture2ChwDecomposition[f][i]-mean)*(concatTexture2ChwDecomposition[f][i]-mean);
//				}
//				
//				std=Math.sqrt(std/(2*n-1));
//				
//				for(int i=0; i<n; i++){
//					concatTexture1ChwDecomposition[f][i]=(concatTexture1ChwDecomposition[f][i]-mean)/std;
//					concatTexture2ChwDecomposition[f][i]=(concatTexture2ChwDecomposition[f][i]-mean)/std;
//				}
//			}

			//fin


			//Compute the mean for each texture responses
			double[][] means = new double[2][features];
			for (int f = 0; f < features; f++) {
				means[0][f] = mean(concatTexture1ChwDecomposition[f]);
				means[1][f] = mean(concatTexture2ChwDecomposition[f]);
			}

			double[][][] concatTextureChwDecomposition=new double[2][features][];
			concatTextureChwDecomposition[0]=new double[features][lenghtImageTexture1];
			concatTextureChwDecomposition[1]=new double[features][lenghtImageTexture2];
			
			for(int i=0; i<features; i++){
				concatTextureChwDecomposition[0][i]=concatTexture1ChwDecomposition[i];
				concatTextureChwDecomposition[1][i]=concatTexture2ChwDecomposition[i];
			}
			
			
			
			//Ajout
//			Matrix S1=new Matrix(new double[features][features]);
//			for (int p = 0; p < lenghtImageTexture1; p++) {
//				
//				double[][] diff=new double[features][1];
//				Matrix diff_mat=new Matrix(new double[lenghtImageTexture1][1]);
//				Matrix mat=new Matrix(new double[features][features]);
//				
//				for (int f = 0; f < features; f++) {
//					diff[f][0]=concatTexture1ChwDecomposition[f][p]-means[0][f];
//				}
//				
//				diff_mat=new Matrix(diff);
//				mat=diff_mat.times(diff_mat.transpose());
//				S1=S1.plus(mat);
//			}
//			
//			Matrix S2=new Matrix(new double[features][features]);
//			for (int p = 0; p < lenghtImageTexture2; p++) {
//				
//				double[][] diff=new double[features][1];
//				Matrix diff_mat=new Matrix(new double[lenghtImageTexture2][1]);
//				Matrix mat=new Matrix(new double[features][features]);
//				
//				for (int f = 0; f < features; f++) {
//					diff[f][0]=concatTexture2ChwDecomposition[f][p]-means[1][f];
//				}
//				
//				diff_mat=new Matrix(diff);
//				mat=diff_mat.times(diff_mat.transpose());
//				S2=S2.plus(mat);
//			}
//			
//			Matrix Sw=S1.plus(S2);
			//
			
			//Within-class scatter matrix
			double[][]  within_class_scatter = computeWithinScatterMatrix(concatTextureChwDecomposition, means);
			
//			//ajout: verification within scatter matrix
//			
//			Matrix scatterClasse1=computeScatterMatrix(concatTextureChwDecomposition[0], means[0]);
//			Matrix scatterClasse2=computeScatterMatrix(concatTextureChwDecomposition[1], means[1]);
//			
//			Matrix within_mat2=scatterClasse1.plus(scatterClasse2);
//			
//			double[][] within=within_mat2.getArrayCopy();
//			
//			//fin ajout

			//Between-class scatter matrix
			double[][] between_class_scatter = computeBetweenScatterMatrix(concatTextureChwDecomposition, means);
			
			//ajout: verification between scatter matrix without the 2*n factor that Yann added in computeBetweenScatterMatrix
//			 double[][] vectMeanClasse1=new double[features][1];
//			 double[][] vectMeanClasse2=new double[features][1];
//			 
//			 for(int i=0; i<features; i++){
//				 vectMeanClasse1[i][0]=means[0][i];
//				 vectMeanClasse2[i][0]=means[1][i];
//			 }
//			
//			Matrix meanClasse1=new Matrix(vectMeanClasse1);
//			Matrix meanClasse2=new Matrix(vectMeanClasse2);
//			
//			Matrix diffMeanBetweenClass=meanClasse2.minus(meanClasse1);
//			Matrix between_mat2=diffMeanBetweenClass.times(diffMeanBetweenClass.transpose());
//			
//			double[][] between=between_mat2.getArrayCopy();
			//fin ajout

			Matrix within_mat = new Matrix(within_class_scatter);
			Matrix between_mat = new Matrix(between_class_scatter);
			

			Matrix mult = (within_mat.inverse()).times(between_mat);
			
			
			//eigenvalue decomposition of S_within^-1 S_between
			EigenvalueDecomposition EVD = new EigenvalueDecomposition(mult);

			//The weights are set as the eigenvector corresponding to the highest eigenvalue
			Matrix eigenVectors = EVD.getV();

			double[] eigenValues = EVD.getRealEigenvalues();

			int maxIndex = 0;
			for (int i = 1; i < eigenValues.length; i++){
				if ((eigenValues[i] > eigenValues[maxIndex])){
					maxIndex = i;
				}
			}

			double[][] eigenVectorsArrayTransposed = eigenVectors.transpose().getArray();

			double[] weights = eigenVectorsArrayTransposed[maxIndex];
			
			//Compute the estimator
//			double[][] weights_2=new double[weights.length][1];
//			double[][] diffMean=new double[means[0].length][1];
//			
//			for(int i=0; i<weights.length; i++){
//				weights_2[i][0]=weights[i];
//				diffMean[i][0]=means[0][i]-means[1][i];
//			}
//			
//			Matrix weights_mat=new Matrix(weights_2);
//
//			Matrix diffMean_mat=new Matrix(diffMean);
//			
//			Matrix num1=(weights_mat.transpose()).times(diffMean_mat);
//			Matrix num=num1.times(num1);
//			Matrix den= (weights_mat.transpose()).times(within_mat.times(weights_mat));			
//			
//			double estimator=num.get(0, 0)/den.get(0, 0);
//			System.out.println("ImageLUTContainer line 1086: "+estimator);
//			String filename="/Users/anais/switchdrive/Projets/Texture_Snakes/My_project/Experiments/estimators.txt";
//			saveEstimator(filename,estimator);
			
			
			//Autre méthode. Donne alpha*weights			
//			double[][] diffMean2=new double[features][1];
//			
//			for (int f = 0; f < features; f++) {
//				diffMean2[f][0]=means[1][f]-means[0][f];
//			}
//			
//			Matrix diffmean_mat=new Matrix(diffMean2);
//			Matrix weightsBis_mat=(within_mat.inverse()).times(diffmean_mat);
//			
//			double[] weightsBis=new double[features];
//			
//			for(int i=0; i<features; i++){
//				weightsBis[i]=weightsBis_mat.get(i, 0);
//			}
//			
//			
//			double b0=weightsBis[0]/weights[0];
//			double b1=weightsBis[1]/weights[1];
//			double b2=weightsBis[2]/weights[2];
//			
//			return weightsBis;
			//
			
			return weights;
		}
	
	// ----------------------------------------------------------------------------

	/**
	 * Compute the within-class covariance matrix
	 * @param responses
	 * @param means
	 * @return
	 */
	private double[][] computeWithinScatterMatrix(double[][][] responses, double[][] means){

		int feature = responses[0].length;

		double[][] within_class_scatter = new double[feature][feature];

		for (int f = 0; f < feature; f++) {
			for (int f2 = 0; f2 < feature; f2++) {
				for (int i = 0; i < 2; i++) {
					within_class_scatter[f][f2] += nCov(responses[i][f], responses[i][f2], means[i][f], means[i][f2]);
				}
			}
		}


		return within_class_scatter;
	}
	
	// ----------------------------------------------------------------------------
	
	/**
	 * Compute a scatter matrix
	 * @param responses
	 * @param means
	 * @return
	 */
	private Matrix computeScatterMatrix(double[][] responses, double[] means){

		int feature = responses.length;

		double[][] vector = new double[feature][1];
		
		for(int i=0; i<feature; i++){
			
			vector[i][0]=0;
		}
		
		Matrix vectorMatrix=new Matrix(vector);
		
		double[][] scatter=new double[feature][feature];
		
		for(int i=0; i<feature; i++){
			for(int j=0; j<feature; j++){
				scatter[i][j]=0;
			}
			
		}
		
		Matrix scatterMatrix = new Matrix(scatter);

		
		int n = responses[0].length;
		for (int i = 0; i < n; i++) {
			for (int f = 0; f < feature; f++) {
				vector[f][0] = responses[f][i]-means[f];
			}
			scatterMatrix = scatterMatrix.plus(vectorMatrix.times(vectorMatrix.transpose()));
		}

		return scatterMatrix;
	}
	
	// ----------------------------------------------------------------------------

	/**
	 * Compute n* covariance between two images
	 * @param img1
	 * @param img2
	 * @param mean1
	 * @param mean2
	 * @return
	 */
	private double nCov(double[] img1, double[] img2, double mean1, double mean2){
		int n = img1.length;
		double covariance = 0;

		for (int i = 0; i < n; i++) {
			covariance += (img1[i] - mean1) * (img2[i] - mean2);
		}

		return covariance;
	}
	
	// ----------------------------------------------------------------------------

	/**
	 * Compute the between-class covariance matrix
	 * @param responses
	 * @param means
	 * @return
	 */
	private double[][] computeBetweenScatterMatrix(double[][][] responses, double[][] means){

		int feature = responses[0].length;

		double[] overall_mean = new double[feature];

		for (int f = 0; f < feature; f++) {
			overall_mean[f] = (means[0][f] + means[1][f])/2.0;
		}

		double[][] between_class_scatter = new double[feature][feature];

		int n = responses[0][0].length;

		for (int f = 0; f < feature; f++) {
			for (int f2 = 0; f2 < feature; f2++) {
				for (int i = 0; i < 2; i++) {
					between_class_scatter[f][f2] += n * (means[i][f] - overall_mean[f]) * (means[i][f2] - overall_mean[f2]);
				}
			}
		}

		return between_class_scatter;
	}

	// ----------------------------------------------------------------------------

	/**
	 * Compute the inverse of the correlation matrix between the chw response maps
	 * @param responses
	 * @return correlation matrix
	 */
	private double[][] computeInverseCorrelationMatrix(double[][][][] responses){
		int nbChannel=sequence_.getSizeC();
		double[][] correlationMatrix = new double[nbChannel*nscales_ * harmonics_.length][nbChannel*nscales_ * harmonics_.length];

		for(int i=0; i<nbChannel; i++){
			for (int sc = 0; sc < nscales_; sc++) {
				for (int h = 0; h < harmonics_.length; h++) {				
					for (int sc2 = 0; sc2 < nscales_; sc2++) {
						for (int h2 = 0; h2 < harmonics_.length; h2++) {
							double corr = correlation(responses[i][sc][h], responses[i][sc2][h2]);
							correlationMatrix[i*nscales_*harmonics_.length+sc * harmonics_.length + h][i*nscales_*harmonics_.length+sc2 * harmonics_.length + h2] = corr;
							correlationMatrix[i*nscales_*harmonics_.length+sc2 * harmonics_.length + h2][i*nscales_*harmonics_.length+sc * harmonics_.length + h] = corr;
						}
					}			
				}
			}
		}


		//Use of the jama library to compute the inverse of the correlation matrix
		Matrix corre = new Matrix(correlationMatrix);
		corre = corre.inverse();
		return corre.getArray();
	}

	// ----------------------------------------------------------------------------

	/**
	 * Compute the correlation between two images
	 * @param img1
	 * @param img2
	 * @return correlation
	 */
	private double correlation(double[] img1, double[] img2){
		int n = img1.length;
		double covariance = 0;

		double mean1 = mean(img1);
		double mean2 = mean(img2);

		for (int i = 0; i < n; i++) {
			covariance += (img1[i] - mean1) * (img2[i] - mean2);
		}
		covariance /= n;

		return covariance /(Math.sqrt(variance(img1) * variance(img2)));
	}
	
	// ----------------------------------------------------------------------------

	/**
	 * Compute the mean of the image
	 * @param img
	 * @return mean
	 */
	public double mean(double[] img){
		double mean = 0;
		for (int i = 0; i < img.length; i++) {
			mean += img[i];

		}
		return mean/img.length;
	}

	// ----------------------------------------------------------------------------
	
	/**
	 * Compute the variance of the image
	 * @param img
	 * @return variance
	 */
	public double variance(double[] img){
		double mean = mean(img);
		double variance = 0;
		for (int i = 0; i < img.length; i++) {
			variance += Math.pow(img[i] - mean, 2);

		}
		return variance/img.length;
	}

	// ----------------------------------------------------------------------------

	/**
	 * Apply an absolute value to an image
	 * @param x image
	 * @return transformed image
	 */
	private double[] absolute(double[] x){
		double[] result = x;
		for (int i = 0; i < x.length; i++) {
			result[i] = Math.abs(x[i]);
		}
		return result;
	}
	
	// ----------------------------------------------------------------------------

	/**
	 * Print 2D array
	 * @param name
	 * @param matrix
	 */
	static void printMatrix(String name, double[][] matrix){
		System.out.println(name + ": ");
		for (int i = 0; i < matrix.length; i++) {
			String s = "[ ";
			for (int j = 0; j < matrix[0].length; j++) {
				s += matrix[i][j] + ", ";
			}
			s += "]";
			System.out.println(s);
		}
	}

	// ----------------------------------------------------------------------------

	/** Generate a double[] from an image load as an ImagePlus */	
	public static double[] ImagePlus2Array1D(ImagePlus imp) {

		int nx = imp.getWidth();
		int ny = imp.getHeight();
		int nxy = nx * ny;
		double[] data = new double[nxy];
		ImageProcessor ip = imp.getProcessor();
		if (ip instanceof ByteProcessor) {
			ByteProcessor bp = (ByteProcessor)ip;
			byte[] pix = (byte[])bp.getPixels();
			for (int k = 0; k < nxy; k++)
				data[k] = pix[k] & 0xFF;
		}
		else if (ip instanceof ShortProcessor) {
			ShortProcessor bp = (ShortProcessor)ip;
			short[] pix = (short[])bp.getPixels();
			for (int k = 0; k < nxy; k++)
				data[k] = pix[k] & 0xFFFF;
		}
		else if (ip instanceof FloatProcessor) {
			FloatProcessor bp = (FloatProcessor)ip;
			float[] pix = (float[])bp.getPixels();
			for (int k = 0; k < nxy; k++)
				data[k] = pix[k];
		}
		else if (ip instanceof ColorProcessor) {
			double r, g, b;
			int c;
			ColorProcessor cp;
			int[] pixels;
			cp = (ColorProcessor) imp.getProcessor();
			pixels = (int[]) cp.getPixels();
			for (int k = 0; k < nxy; k++) {
				c = pixels[k];
				r = (c & 0xFF0000) >> 16;
			g = (c & 0xFF00) >> 8;
			b = ((c & 0xFF));
			data[k] = (r + g + b) / 3.0;
			}

		}

		return data;
	}
	
	/**
	 * Convert a 2D to 1D array
	 * @param image
	 * @return
	 */
	private double[] image2Dto1D(double[][] image){
		double[] result = new double[image.length * image[0].length];
		int k =0;
		for (int i = 0; i < image[0].length; i++) {
			for (int j = 0; j < image.length; j++) {
				result[k] = image[j][i];
				k++;
			}
		}
		return result;
	}

	// ----------------------------------------------------------------------------

	/** If <code>true</code>, the LUTs are consistent with the input parameters. */
	public boolean isLUTUpToDate() {
		return isLUTUpToDate_;
	}
	
	public void setLUTUpToDate(boolean upToDate){
		isLUTUpToDate_=upToDate;
	}

	// ----------------------------------------------------------------------------

	/** Returns a string with all the information of the LUT. */
	@Override
	public String toString() {
		return new String("[LUT container: "
				+ ", sequence title = " + sequence_.getName()
				+ ", channel number = " + channelNumber_
				+ ", LUT up to date = " + isLUTUpToDate_ + "]");
	}

	// ============================================================================
	// SETTERS AND GETTERS

	/** Sets the active channel. */
	public void setChannelNumber(int channelNumber) {
		isLUTUpToDate_ = false;
		channelNumber_ = channelNumber;
	}

	// ----------------------------------------------------------------------------

	/** Returns the active channel. */
	public int getChannelNumber() {
		return channelNumber_;
	}

	// ----------------------------------------------------------------------------

	public void setScales(int scales) {
		isLUTUpToDate_ = false;
		nscales_ = scales;
	}
	
	// ----------------------------------------------------------------------------

	public void setTexture1(IcyBufferedImage texture1, ROI2DRectangle ROITexture1) {
		isLUTUpToDate_ = false;
		
		int nbChannel=sequence_.getSizeC();
		texture1_=new double[nbChannel][];
		
		IcyBufferedImage[] inputCopy =new IcyBufferedImage[nbChannel];
		IcyBufferedImage[] inputRGBImage =new IcyBufferedImage[nbChannel];
		
		for(int i=0; i<nbChannel; i++){
			inputRGBImage[i]=texture1.getImage(i);
			
			//To display each channel
//			Sequence sequence = new Sequence("Channel " + i);
//			sequence.addImage(inputRGBImage[i]);
//			Icy.getMainInterface().addSequence(sequence);
			
			inputCopy[i] = IcyBufferedImageUtil.convertToType(
					inputRGBImage[i], DataType.DOUBLE, false);
			
			texture1_[i]=inputCopy[i].getDataXYAsDouble(0);
		}
		
		ROITexture1_=ROITexture1;
	}
	
	// ----------------------------------------------------------------------------
	public void setTexture2(IcyBufferedImage texture2, ROI2DRectangle ROITexture2) {
		isLUTUpToDate_ = false;
		
		int nbChannel=sequence_.getSizeC();
		texture2_=new double[nbChannel][];
		
		IcyBufferedImage[] inputCopy =new IcyBufferedImage[nbChannel];
		IcyBufferedImage[] inputRGBImage =new IcyBufferedImage[nbChannel];
		
		for(int i=0; i<nbChannel; i++){
			
			inputRGBImage[i]=texture2.getImage(i);
			
			//To display each channel
//			Sequence sequence = new Sequence("Channel " + i);
//			sequence.addImage(inputRGBImage[i]);
//			Icy.getMainInterface().addSequence(sequence);
			
			inputCopy[i] = IcyBufferedImageUtil.convertToType(
					inputRGBImage[i], DataType.DOUBLE, false);
			
			texture2_[i]=inputCopy[i].getDataXYAsDouble(0);
		}
	
		ROITexture2_=ROITexture2;
	
	}
	
	// ----------------------------------------------------------------------------

	public void setTextureDimension(int width, int height) {
		isLUTUpToDate_ = false;
		this.textureWidth_=width;
		this.textureHeight_=height;
	}
	
	// ----------------------------------------------------------------------------

	public int getTextureWidth() {
		return this.textureWidth_;
	}
	
	// ----------------------------------------------------------------------------

	public int getTextureHeight() {
		return this.textureHeight_;
	}
	
	// ----------------------------------------------------------------------------
	public double[][] getTexture1() {
		return texture1_;
	}

	// ----------------------------------------------------------------------------
	public double[][] getTexture2() {
		return texture2_;
	}
	
	// ----------------------------------------------------------------------------
	public ROI2DRectangle getROITexture1() {
		return ROITexture1_;
	}
	
	// ----------------------------------------------------------------------------
	public ROI2DRectangle getROITexture2() {
		return ROITexture2_;
	}

	// ----------------------------------------------------------------------------

	/** Sets the standard deviation of the Gaussian kernel used for smoothing. */
	public void setHarmonics(int numHarmonics) {
		isLUTUpToDate_ = false;
		int[] h = new int[numHarmonics];
		for (int i = 0; i < numHarmonics; i++) {
			h[i] = i;
		}
		harmonics_ = h;
	}

	// ----------------------------------------------------------------------------

	/**
	 * Returns the ammount of scales for the chw.
	 */
	public int getScale(){
		return nscales_;
	}

	// ----------------------------------------------------------------------------

	/**
	 * Returns harmonics used for chw
	 */
	public int[] getHarmonics(){
		return harmonics_;
	}

	/**
	 * Returns the correlation matrix between the responses
	 */
	public double[][] getCorrelationMatrix_() {
		return correlationMatrix_;
	}

	/**
	 * Returns the fisher weight
	 */
	public double[] getFisherWeights_() {
		return fisherWeights_;
	}


	// ----------------------------------------------------------------------------

	/** Sets the active image from where the data is extracted. */
	public void setSequence(Sequence sequence) {
		isLUTUpToDate_ = false;
		sequence_ = sequence;
	}

	// ----------------------------------------------------------------------------

	/** Returns the image from where the image data is extracted. */
	public Sequence getOriginalSequence() {
		return sequence_;
	}

	// ----------------------------------------------------------------------------

	/** Returns the with in pixels of the image. */
	public int getImageWidth() {
		return width_;
	}

	// ----------------------------------------------------------------------------

	/** Returns the height in pixels of the image. */
	public int getImageHeight() {
		return height_;
	}

	// ----------------------------------------------------------------------------

	/**
	 * Returns a one dimensional array containing the pixel values of the
	 * original image of the active channel.
	 */
	public double[][] getImageDataArray() {
		return imageDataArray_;
	}

	// ----------------------------------------------------------------------------

	/**
	 * Returns a one dimensional array containing the pixel values of the
	 * original image after preintegration of the active channel.
	 */
	public double[][] getPreintegratedImageDataArray() {
		return preintegratedImageDataArray_;
	}
	
	public double[][] getExtendedPreintegratedImageDataArray() {
		return extendedPreintegratedImageDataArray_;
	}

	// ----------------------------------------------------------------------------

	/**
	 * Returns a three dimensional array containing the pixel values of the decomposition
	 * of the image through the circularl harmonic wavelets for each scale and harmonic
	 */
	public double[][][][] getChwDecomposition_() {
		return chwDecomposition_;
	}
	
	// ----------------------------------------------------------------------------

	/**
	 * Returns a three dimensional array containing the pixel values of the decomposition
	 * of the image through the circularl harmonic wavelets for each scale and harmonic
	 * after preintegration
	 */
	public double[][][][] getPreintegratedChwDecomposition_() {
		return preintegratedChwDecomposition_;
	}
	
	// ----------------------------------------------------------------------------

	public double[][][][] getExtendedPreintegratedChwDecomposition_() {
		return extendedPreintegratedChwDecomposition_;
	}
	
	// ----------------------------------------------------------------------------

	public int getNumberChannel() {
		return sequence_.getSizeC();
	}
	
	// ----------------------------------------------------------------------------

	/**
	 * Remove all ROIs of the sequence
	 */
	public void removeROI(){
		ImageLUTContainer.sequence_.removeAllROI();
	}
}