/*******************************************************************************
 * Copyright (c) 2012-2013 Biomedical Image Group (BIG), EPFL, Switzerland.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Public License v3.0
 * which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/gpl.html
 * 
 * Contributors:
 *     Ricard Delgado-Gonzalo (ricard.delgado@gmail.com)
 *     Nicolas Chenouard (nicolas.chenouard@gmail.com)
 *     Philippe Th&#233;venaz (philippe.thevenaz@epfl.ch)
 *     Emrah Bostan (emrah.bostan@gmail.com)
 *     Ulugbek S. Kamilov (kamilov@gmail.com)
 *     Ramtin Madani (ramtin_madani@yahoo.com)
 *     Masih Nilchian (masih_n85@yahoo.com)
 *     C&#233;dric Vonesch (cedric.vonesch@epfl.ch)
 *     Virginie Uhlmann (virginie.uhlmann@epfl.ch)
 *     Cl&#233;ment Marti (clement.marti@epfl.ch)
 *     Julien Jacquemot (julien.jacquemot@epfl.ch)
 ******************************************************************************/
package plugins.big.texturesnake.snake;

import icy.util.XMLUtil;
import plugins.big.texturesnake.core.Settings;

import org.w3c.dom.Element;
import org.w3c.dom.Node;

/**
 * Class that wraps the parameters of the E-Snake.
 * 
 * @version October 27, 2014
 * 
 * @author Ricard Delgado-Gonzalo (ricard.delgado@gmail.com)
 * @author Daniel Schmitter (daniel.schmitter@epfl.ch)
 */
public class ESnakeParameters {


	/** Number of spline vector coefficients. */
	private int M_ = Settings.M_DEFAULT;
	/** Indicates the energy function of the snake. */
	private ESnakeEnergyType energyType_ = Settings.ENERGY_TYPE_DEFAULT;
	/**
	 * Maximum number of iterations allowed when the <code>immortal</code> flag
	 * is <code>false</code>.
	 */
	private int maxLife_ = Settings.MAX_LIFE_DEFAULT;
	/**
	 * If <code>true</code> indicates that the snake will keep iterating till
	 * the optimizer decides so.
	 */
	private boolean immortal_ = Settings.IMMORTAL_DEFAULT;

	// ----------------------------------------------------------------------------
	// XML TAG FIELDS

	/**
	 * Label of the XML tag informing about the brightness of the target to
	 * segment (brighter or darker than the background).
	 */
	public static final String ID_DETECT_TYPE = "detectType";
	/**
	 * Label of the XML tag containing the number of control points of the
	 * snake.
	 */
	public static final String ID_M = "M";
	/**
	 * Label of the XML tag informing about the type of energy the snake uses
	 * (contour, region or mixture).
	 */
	public static final String ID_ENERGY_TYPE = "energyType";
	/** Label of the XML tag containing the maximum number of iterations. */
	public static final String ID_MAX_LIFE = "maxLife";
	/**
	 * Label of the XML tag informing if the number of iterations of the snake
	 * is limited or not.
	 */
	public static final String ID_IMMORTAL = "immortal";

	// ============================================================================
	// PUBLIC METHODS

	/** Default constructor. */
	public ESnakeParameters() {
	}

	// ----------------------------------------------------------------------------

	/** Constructor. */
	public ESnakeParameters(int M,
			ESnakeEnergyType energyType, int maxLife, boolean immortal) {
		setMaxLife(maxLife);
		setM(M);
		setImmortal(immortal);
		setEnergyType(energyType);
	}

	// ----------------------------------------------------------------------------

	@Override
	public String toString() {
		return new String("[Snake parameters: " + ID_M + " " + M_ + ID_ENERGY_TYPE + " "
				+ energyType_ + " "+ ID_MAX_LIFE + " "
				+ maxLife_ + ID_IMMORTAL + " " + immortal_ + "]");
	}

	// ----------------------------------------------------------------------------

	public void loadFromToXML(Node node) {

		Element energyTypeAsStringElement = XMLUtil.getElement(node,
				ID_ENERGY_TYPE);
		if (energyTypeAsStringElement != null) {
			String energyTypeAsString = energyTypeAsStringElement
					.getTextContent();
			if (energyTypeAsString.compareToIgnoreCase("TEXTURE_COR") == 0) {
				setEnergyType(ESnakeEnergyType.TEXTURE_COR);
			} else if (energyTypeAsString.compareToIgnoreCase("TEXTURE_FISH") == 0) {
				setEnergyType(ESnakeEnergyType.TEXTURE_FISH);
			} else {
				setEnergyType(Settings.ENERGY_TYPE_DEFAULT);
			}
		} else {
			setEnergyType(Settings.ENERGY_TYPE_DEFAULT);
		}



		setMaxLife(XMLUtil.getElementIntValue(node, ID_MAX_LIFE,
				Settings.MAX_LIFE_DEFAULT));
		
		setM(XMLUtil.getElementIntValue(node, ID_M, Settings.M_DEFAULT));

		setImmortal(XMLUtil.getElementBooleanValue(node, ID_IMMORTAL,
				Settings.IMMORTAL_DEFAULT));
	}

	// ----------------------------------------------------------------------------

	public void saveToXML(Element node) {
		XMLUtil.setElementIntValue(node, ESnakeParameters.ID_M, M_);
		switch (energyType_) {
		case TEXTURE_COR:
			XMLUtil.setElementValue(node, ESnakeParameters.ID_ENERGY_TYPE,
					"TEXTURE_COR");
			break;
		case TEXTURE_FISH:
			XMLUtil.setElementValue(node, ESnakeParameters.ID_ENERGY_TYPE,
					"TEXTURE_FISH");
			break;
		}

		XMLUtil.setElementDoubleValue(node, ESnakeParameters.ID_MAX_LIFE,
				maxLife_);
		
		XMLUtil.setElementBooleanValue(node, ESnakeParameters.ID_IMMORTAL,
				immortal_);
	}

	// ============================================================================
	// GETTERS AND SETTERS


	public int getM() {
		return M_;
	}

	// ----------------------------------------------------------------------------

	public void setM(int M) {
		M_ = M;
	}

	// ----------------------------------------------------------------------------

	public ESnakeEnergyType getEnergyType() {
		return energyType_;
	}

	// ----------------------------------------------------------------------------

	public void setEnergyType(ESnakeEnergyType energyType) {
		energyType_ = energyType;
	}

	// ----------------------------------------------------------------------------

	public int getMaxLife() {
		return maxLife_;
	}

	// ----------------------------------------------------------------------------

	public void setMaxLife(int maxLife) {
		maxLife_ = maxLife;
	}

	// ----------------------------------------------------------------------------

	public boolean isImmortal() {
		return immortal_;
	}

	// ----------------------------------------------------------------------------

	public void setImmortal(boolean immortal) {
		immortal_ = immortal;
	}
}
